import _ from 'lodash'
import { createAnnotation, deleteNode, TOCProvider as TocProvider } from 'substance'

class CommentsProvider extends TocProvider {
  constructor (document, props) {
    super(document, props)
    const self = this

    self.activeEntry = null
    props.documentSession.on('didUpdate', this.updateActiveEntry, this)
  }

  //
  // entries

  computeEntries () {
    const comments = this.getComments()
    const entries = this.sortNodes(comments)
    return entries
  }

  reComputeEntries () {
    this.entries = this.computeEntries()
    this.emit('comments:updated')
  }

  //
  // active entry

  getActiveEntry () {
    return this.activeEntry
  }

  setActiveEntry (id) {
    this.activeEntry = id
    this.emit('comments:updated')
  }

  removeActiveEntry () {
    this.activeEntry = null
    this.emit('comments:updated')
  }

  updateActiveEntry () {
    // if the new selection is not a comment, remove the active entry (if one)
    // if the new selection is a comment, make it activeEntry
    const activeEntry = this.getActiveEntry()
    const commentState = this.getCommentState()
    const onComment = commentState.active  // is current selection on a comment

    if (!onComment && !activeEntry) return

    if (!onComment && activeEntry) {
      this.removeEmptyComments()
      this.removeActiveEntry()
      return
    }

    const activeComment = this.getActiveComment()
    if (activeEntry === activeComment.id) return
    this.setActiveEntry(activeComment.id)
  }

  //
  // resolved comments

  resolveComment (id) {
    const self = this

    const ds = this.config.documentSession
    const doc = ds.getDocument()
    const commentNode = doc.get(id)

    const path = commentNode.path
    const startOffset = commentNode.startOffset
    const endOffset = commentNode.endOffset

    const sel = ds.createSelection(path, startOffset, endOffset)

    const resolvedNodeData = {
      selection: sel,
      node: { type: 'resolved-comment' }
    }

    // create resolved comment annotation on the same selection the
    // comment was on and remove existing comment
    ds.transaction(function (tx, args) {
      const annotation = createAnnotation(doc, resolvedNodeData)
      const resolvedCommentId = annotation.node.id
      self.markCommentAsResolved(id, resolvedCommentId)
    })

    this.deleteCommentNode(id)
    this.reComputeEntries()
  }

  //
  // focus text area for a comment reply

  focusTextArea (id) {
    setTimeout(function () {
      var textarea = document.getElementById(id)
      if (textarea) {
        textarea.focus()
      }
    }, 10)
  }

  //
  // remove comments from pane if the user wrote nothing in them

  removeEmptyComments () {
    const self = this

    const comments = self.getComments()
    const commentsContent = self.config.comments

    _.each(comments, function (value, key) {
      if (commentsContent && !commentsContent[key]) {
        self.deleteCommentNode(key)
      }
    })
  }

  //
  //
  // helpers
  //
  //

  deleteCommentNode (id) {
    const surface = this.getSurface()

    surface.transaction(function (tx, args) {
      deleteNode(tx, { nodeId: id })
    })
  }

  getActiveComment () {
    const selectionState = this.config.documentSession.getSelectionState()
    const commentAnnotations = selectionState.getAnnotationsForType('comment')

    // TODO -- what happens with multiple comments on the same cursor position?
    return commentAnnotations[0]
  }

  getComments () {
    const doc = this.getDocument()
    const nodes = doc.getNodes()

    // get all notes from the document
    const comments = _.pickBy(nodes, function (value, key) {
      return value.type === 'comment'
    })

    return comments
  }

  getCommentState () {
    const { commandManager } = this.config
    const commandStates = commandManager.getCommandStates()
    return commandStates.comment
  }

  getSurface () {
    const { containerId, surfaceManager } = this.config
    return surfaceManager.getSurface(containerId)
  }

  // TODO -- do I need to override this?
  handleDocumentChange (change) {
    var doc = this.getDocument()
    var needsUpdate = false
    var tocTypes = this.constructor.tocTypes

    for (var i = 0; i < change.ops.length; i++) {
      var op = change.ops[i]
      var nodeType
      if (op.isCreate() || op.isDelete()) {
        var nodeData = op.getValue()
        nodeType = nodeData.type
        if (_.includes(tocTypes, nodeType)) {
          needsUpdate = true
          break
        }
      } else {
        var id = op.path[0]
        var node = doc.get(id)
        if (node && _.includes(tocTypes, node.type)) {
          needsUpdate = true
          break
        }
      }
    }

    if (needsUpdate) {
      // need a timeout here, to make sure that the updated doc has rendered
      // the annotations, so that the comment box list can be aligned with them
      const self = this
      setTimeout(function () {
        self.reComputeEntries()
      })
    }
  }

  /*
  When a comment gets resolved the comment on the fragment gets a resolved property.
  The resolved property is the id of the new resolved comment annotation.
  If the user does not save the document after resolving, the id for the comment is still there, so it will display correctly.
  If the user resolves the comment, does not save and then resolves the same comment again, the resolve property will simply be overwritten.
  If the document is saved, the comment id will not exist in the document any more, so it will not be rendered.
  If the now resolved comment needs to be unresolved, the original contents can easily be found by looking for the object with the resolved comment's id in the resolved property.
  */
  markCommentAsResolved (commentId, resolvedId) {
    const comments = this.config.comments
    const update = this.config.updateComments
    const id = commentId

    if (!comments[id]) return
    comments[id].resolved = resolvedId
    update(comments)
  }

  sortNodes (nodes) {
    var comments = _.clone(nodes)
    var doc = this.getDocument()
    var container = doc.get('body')

    // sort notes by
    //   the index of the containing block
    //   their position within that block
    comments = _.map(comments, function (comment) {
      const blockId = comment.path[0]
      const blockPosition = container.getPosition(blockId)
      const nodePosition = comment.startOffset

      return {
        id: comment.id,
        blockPosition: blockPosition,
        nodePosition: nodePosition,
        node: comment
      }
    })

    return _.sortBy(comments, ['blockPosition', 'nodePosition'])
  }
}

CommentsProvider.tocTypes = ['comment']

export default CommentsProvider
